package it.unibz.inf.completeness;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import com.hp.hpl.jena.graph.Triple;
import com.hp.hpl.jena.query.Dataset;
import com.hp.hpl.jena.query.Query;
import com.hp.hpl.jena.query.QueryExecution;
import com.hp.hpl.jena.query.QueryExecutionFactory;
import com.hp.hpl.jena.query.ResultSet;
import com.hp.hpl.jena.query.ResultSetFormatter;
import com.hp.hpl.jena.rdf.model.Model;
import com.hp.hpl.jena.rdf.model.ModelFactory;
import com.hp.hpl.jena.rdf.model.Statement;
import com.hp.hpl.jena.rdf.model.StmtIterator;
import com.hp.hpl.jena.sparql.core.BasicPattern;
import com.hp.hpl.jena.sparql.core.Substitute;
import com.hp.hpl.jena.sparql.core.Var;
import com.hp.hpl.jena.sparql.engine.binding.Binding;
import com.hp.hpl.jena.sparql.engine.binding.BindingFactory;
import com.hp.hpl.jena.sparql.engine.binding.BindingMap;
import com.hp.hpl.jena.sparql.engine.binding.BindingProject;
import com.hp.hpl.jena.sparql.engine.binding.BindingUtils;
import com.hp.hpl.jena.tdb.TDB;
import com.hp.hpl.jena.tdb.TDBFactory;

import it.unibz.inf.semweb.QueryToolkit;
import it.unibz.inf.semweb.QueryWithNegation;

// TDB-backed completeness reasoner
public class CompletenessReasoner {
	
	private Dataset ds;
	
	// connect to a TDB dataset
	public CompletenessReasoner(String path) {
		ds = TDBFactory.createDataset(path);
	}

	// TDB.sync flushes the changes on the dataset to the disk to avoid corrupted TDB store
	// credit: http://stackoverflow.com/questions/15640226/store-data-in-jena-tdb-and-use-reasoning
	public void close() {
		TDB.sync(ds);
		ds.close();
	}

	// direct access must be done carefully as otherwise it may corrupt the dataset
	public Dataset getDataset() {
		return ds;
	}
	
	public void addToGraph(Model graphToAdd) {
		Model model = ds.getDefaultModel();
		model.add(graphToAdd);
	}

	public void removeFromGraph(Model graphToRemove) {
		Model model = ds.getDefaultModel();
		model.remove(graphToRemove);
	}

	public List<Binding> evalSELECT(Query q) {
		List<Binding> mappings = new ArrayList<Binding>();
		Model model = ds.getDefaultModel();
		QueryExecution qExec = QueryExecutionFactory.create(q, model);
		ResultSet rs = qExec.execSelect();
		while(rs.hasNext()) {
			Binding thisMapping = rs.nextBinding();
			mappings.add(thisMapping);
		}
		return mappings;
	}

	public void evalSELECTAndDisplay(Query q) {
		Model model = ds.getDefaultModel();
		QueryExecution qExec = QueryExecutionFactory.create(q, model);
		System.out.println(ResultSetFormatter.asText(qExec.execSelect()));
	}

	public void evalSELECTAndConsume(Query q) {
		Model model = ds.getDefaultModel();
		QueryExecution qExec = QueryExecutionFactory.create(q, model);
		ResultSetFormatter.consume(qExec.execSelect());
	}
	
	public boolean contains(Model m) {
		Model model = ds.getDefaultModel();
		return model.containsAll(m);
	}

	public boolean contains(Statement stmt) {
		Model model = ds.getDefaultModel();
		return model.contains(stmt);
	}
	
	////////////////////////////////////////////////
	// method for naive completeness checking without graphs (aka schema reasoning)
	////////////////////////////////////////////////

	public boolean checkSchemaCompletenessNaive(BasicPattern bgp, List<CompletenessStatement> statements) {
		Model prototypicalGraph = QueryToolkit.freeze(bgp);
		Model availGraph = ModelFactory.createDefaultModel();
  		for(CompletenessStatement statement : statements) {
  			Query q = statement.getConstructQuery();
  			QueryExecution qexec = QueryExecutionFactory.create(q, prototypicalGraph);
  			availGraph.add(qexec.execConstruct());
  		}
		return availGraph.containsAll(prototypicalGraph);
	}

	public boolean checkConditionalSchemaCompletenessNaive(Model negPrototypicalGraph, Model posAndNegPrototypicalGraph, List<CompletenessStatement> statements) {
		Model availGraph = ModelFactory.createDefaultModel();
  		for(CompletenessStatement statement : statements) {
  			Query q = statement.getConstructQuery();
  			QueryExecution qexec = QueryExecutionFactory.create(q, posAndNegPrototypicalGraph);
  			availGraph.add(qexec.execConstruct());
  		}
		return availGraph.containsAll(negPrototypicalGraph);
	}

	////////////////////////////////////////////////
	// method for optimized completeness checking without graphs (aka schema reasoning)
	////////////////////////////////////////////////
	
	public boolean checkSchemaCompletenessOptimized(BasicPattern bgp, ConstantMap constantMap) {
		List<String> queryConstants = ConstantMap.getUniqueConstantList(bgp);
		List<CompletenessStatement> relevantCSList = constantMap.search(queryConstants);		
		return checkSchemaCompletenessNaive(bgp, relevantCSList);
	}
	
	public boolean checkConditionalSchemaCompletenessOptimized(Model negPrototypicalGraph, List<Triple> posAndNegBGP, Model posAndNegPrototypicalGraph, ConstantMap constantMap) {
		List<String> queryConstants = ConstantMap.getUniqueConstantList(BasicPattern.wrap(posAndNegBGP));
		List<CompletenessStatement> relevantCSList = constantMap.search(queryConstants);
		
		// DEBUG
//		System.out.println("Size of relevant CSs: " + relevantCSList.size());
		
		return checkConditionalSchemaCompletenessNaive(negPrototypicalGraph, posAndNegPrototypicalGraph, relevantCSList);
	}

	////////////////////////////////////////////////
	// methods for naive completeness checking
	////////////////////////////////////////////////
	
	public boolean checkCompletenessNaive(BasicPattern bgp, List<CompletenessStatement> statements) {
		try {
			saturateAndCheckContainmentNaive(bgp, statements);
		}
		catch(IncompletenessException ie) {
			return false;
		}
		return true;
	}

	private void saturateAndCheckContainmentNaive(BasicPattern bgp,
			List<CompletenessStatement> statements) throws IncompletenessException {
		
		List<BasicPattern> Pworking = new ArrayList<BasicPattern>();
		Pworking.add(bgp);
		
		while(!(Pworking.isEmpty())) {
			BasicPattern currentBGP = Pworking.get(0);
			Pworking.remove(0);
			Set<BasicPattern> Pequiv = epgNaive(currentBGP, statements);

			// if saturated then ...
			if(Pequiv.size() != 0 && Pequiv.iterator().next().toString().compareTo(currentBGP.toString()) == 0) {
				
				// check containment
				if(!(this.contains(QueryToolkit.freeze(currentBGP))))
					throw new IncompletenessException();
				
			}
			// else if Pequiv is not empty, add the instantiations to Pworking
			else if(Pequiv.size() != 0) {
				
				// DFS-based traversal
				List<BasicPattern> Ptemp = new ArrayList<BasicPattern>(Pworking);
				Pworking = new ArrayList<BasicPattern>();
				Pworking.addAll(Pequiv);
				Pworking.addAll(Ptemp);
				
			} // otherwise, do nothing -- optimization rule: emptiness skip
		}
		
	}

	public Set<BasicPattern> epgNaive(BasicPattern bgp, List<CompletenessStatement> statements) {
		Set<BasicPattern> bgps = new HashSet<BasicPattern>();
		
		BasicPattern crucPart = this.getCrucialPartNaive(bgp, statements);

		// optimization rule: completeness skip
		// NOTE: this optimization rule assumes that BGPs contain no duplicates
		if(crucPart.size() == bgp.size()) {
			return bgps;
		}
		
		if(crucPart.size() == 0) {
			bgps.add(bgp);
			return bgps;
		}
		
		List<Triple> bgpToInstantiate = new ArrayList<Triple>(bgp.getList());
		
		// optimization rule: BGP length reduction
		bgpToInstantiate.removeAll(crucPart.getList());
		
		// optimization rule: branching reduction
		bgps = this.instantiateNaive(crucPart, BasicPattern.wrap(bgpToInstantiate));
		
		return bgps;
	}

	public BasicPattern getCrucialPartNaive(BasicPattern bgp, List<CompletenessStatement> statements) {
		Model crucPart = ModelFactory.createDefaultModel();
		
		Model prototypicalGraph = QueryToolkit.freeze(bgp);
		
		Model existingPrototypicalGraph = ModelFactory.createDefaultModel(); // we detect which parts of the proptotypical graph already exist in the graph, if yes, we do not remove them later
		StmtIterator stmtIt = prototypicalGraph.listStatements();
		while(stmtIt.hasNext()) {
			Statement stmt = stmtIt.next();
			if(contains(stmt))
				existingPrototypicalGraph.add(stmt);
		}
		
		addToGraph(prototypicalGraph);
		
		Model model = ds.getDefaultModel();			
		for(CompletenessStatement statement : statements) {
			QueryExecution qExec = QueryExecutionFactory.create(statement.getConstructQuery(), model);
			crucPart.add(prototypicalGraph.intersection(qExec.execConstruct()));
		}

		prototypicalGraph.remove(existingPrototypicalGraph);
		removeFromGraph(prototypicalGraph);
		return QueryToolkit.defreeze(crucPart);
	}

	// performance-wise: sufficient to consider only the part (or variables) of mappings relevant for the BGP to instantiate
	private Set<BasicPattern> instantiateNaive(BasicPattern crucPart,
			BasicPattern bgpToInstantiate) {
		Set<BasicPattern> bgps = new HashSet<BasicPattern>();
		Model model = null;
		model = ds.getDefaultModel();
		QueryExecution qExec = QueryExecutionFactory.create(QueryToolkit.bgpToSelectQuery(crucPart), model);
		ResultSet queryResult = qExec.execSelect();
		
		// instantiations
		while(queryResult.hasNext()) {
			bgps.add(Substitute.substitute(bgpToInstantiate, queryResult.nextBinding()));
		}
		
		return bgps;
	}

	////////////////////////////////////////////////
	// methods for optimized completeness checking using completeness templates data structure
	////////////////////////////////////////////////
	
	// as a difference to the naive completeness checking, this template-based checking is recursive
	public boolean checkCompletenessOptimized(BasicPattern bgp, HashMap<String, List<CompletenessTemplate>> templateMap) {
		
		// DEBUG
//		System.out.println("[[[\nTempMap -> " + templateMap.values().iterator().next().get(0).bindings + "\n]]]");
		
		try {
			saturateAndCheckContainmentOptimized(bgp, templateMap, new HashSet<String>());
		}
		catch(IncompletenessException ie) {
			return false;
		}
		return true;
	}

	private void saturateAndCheckContainmentOptimized(BasicPattern bgp,
			HashMap<String, List<CompletenessTemplate>> templateMap,
			Set<String> checkedSignatures) throws IncompletenessException {

		// DEBUG
//		System.out.println("BEFORE EPG OPTIMIZED");
		
		Set<BasicPattern> Pequiv = epgOptimized(bgp, templateMap, checkedSignatures); // we need not check again the parts that are known useless for completeness checking
		
		// DEBUG
//		System.out.println("AFTER EPG OPTIMIZED");
//		System.out.println(bgp + " ### " + Pequiv);
		
		// base case
		if(Pequiv.size() != 0 && Pequiv.iterator().next().toString().compareTo(bgp.toString()) == 0) {
			if(!(this.contains(QueryToolkit.freeze(bgp))))
				throw new IncompletenessException();
		}
		else if(Pequiv.size() != 0) {
			
			Set<String> newCheckedSignatures = SignatureFactory.getSignatureWithGeneralization(intersection(bgp, Pequiv.iterator().next()));
			Iterator<BasicPattern> PequivIt = Pequiv.iterator();
			while(PequivIt.hasNext()) {
				BasicPattern thisBGP = PequivIt.next();
				saturateAndCheckContainmentOptimized(thisBGP, templateMap, newCheckedSignatures);
			}
		}
				
	}
	
	// auxiliary method
	public static List<Triple> intersection(BasicPattern bgp1, BasicPattern bgp2) {		
		List<Triple> bgp1AsList = bgp1.getList();
		List<Triple> bgp2AsList = bgp2.getList();
		
		List<Triple> intersectionList = new ArrayList<Triple>();
		for(Triple t : bgp2AsList) {
			if(bgp1AsList.contains(t))
				intersectionList.add(t);
		}
		
		return intersectionList;
	}
	
	private Set<BasicPattern> epgOptimized(BasicPattern currentBGP,
			HashMap<String, List<CompletenessTemplate>> templateMap,
			Set<String> checkedSignatures) {
		Set<BasicPattern> epgSet = new HashSet<BasicPattern>();
		
		// DEBUG
//		System.out.println("[[[\nTempMap -> " + templateMap.values().iterator().next().get(0).bindings + "\n]]]");
		
		Set<CompletenessTemplate> partiallyMatchedTemplates = partialMatch(currentBGP, checkedSignatures, templateMap);
		
//		BasicPattern crucPart = QueryToolkit.defreeze(this.getCrucialPartOptimized(currentBGP, partiallyMatchedTemplates));
		
		// super optimized as per Werner's idea for subset generation and better separation of evaluation over prototypical graph vs data graph
		BasicPattern crucPart = QueryToolkit.defreeze(this.getCrucialPartSuperOptimized(currentBGP, partiallyMatchedTemplates));
				
		// DEBUG
//		System.out.println("PMATCH: " + partiallyMatchedTemplates + "\n");
//		System.out.println("CRUC: " + crucPart + "\n");
//		this.close();
//		System.exit(0);
		
		// optimization rule: completeness skip
		if(crucPart.size() == currentBGP.size()) {
			return epgSet;
		}
		
		if(crucPart.size() == 0) {
			epgSet.add(currentBGP);
			return epgSet;
		}
		
		List<Triple> bgpToInstantiate = new ArrayList<Triple>(currentBGP.getList());
		
		// optimization rule: BGP length reduction
		bgpToInstantiate.removeAll(crucPart.getList());
				
		// optimization rule: branching reduction
		epgSet = this.instantiateNaive(crucPart, BasicPattern.wrap(bgpToInstantiate));
		
		// DEBUG
//		System.out.println("EPG SET: " + epgSet + "\n");
		
		return epgSet;
	}
	
	// optimization rule: focused partial matching
	public Set<CompletenessTemplate> partialMatch(BasicPattern bgp, Set<String> checkedSignatures, HashMap<String, List<CompletenessTemplate>> signatureMap) {
		Set<String> uncheckedSignatures = SignatureFactory.getUncheckedSignatures(bgp.getList(), checkedSignatures);		
		return SignatureFactory.getCompletenessTemplatesFromSignatures(uncheckedSignatures, signatureMap);
	}

	public Model getCrucialPartOptimized(BasicPattern bgpToCheck, Set<CompletenessTemplate> templates) {
	
		Model prototypicalGraph = QueryToolkit.freeze(bgpToCheck);
		Model crucPart = ModelFactory.createDefaultModel();
		
		Model existingPrototypicalGraph = ModelFactory.createDefaultModel();
		StmtIterator stmtIt = prototypicalGraph.listStatements();
		while(stmtIt.hasNext()) {
			Statement stmt = stmtIt.next();
			if(contains(stmt))
				existingPrototypicalGraph.add(stmt);
		}
		
		this.addToGraph(prototypicalGraph);
		
		// DEBUG
//		System.out.println("[[[getCrucialPartOptimized for -> " + prototypicalGraph + "\n]]]");
		
		Model model = null;	
		model = ds.getDefaultModel();
		Iterator<CompletenessTemplate> templatesIt = templates.iterator();
		while(templatesIt.hasNext()) { // for each template, do ...
			CompletenessTemplate localTemplate = templatesIt.next();
			List<Var> localTemplateVars = localTemplate.getPseudoVars();
			BasicPattern localTemplateBGP = localTemplate.getBGP();
			
			// partial query evaluation
			for(Triple t : localTemplateBGP) {
				List<Triple> tAsList = new ArrayList<Triple>();
				tAsList.add(t);
				List<Triple> reducedLocalTemplateBGPAsList = new ArrayList<Triple>(localTemplateBGP.getList());
				reducedLocalTemplateBGPAsList.remove(t);
				BasicPattern reducedLocalTemplateBGP = BasicPattern.wrap(reducedLocalTemplateBGPAsList);
				
				// evaluate each triple pattern over the prototypical graph
				QueryExecution tripleQueryExec = QueryExecutionFactory.create(QueryToolkit.bgpToSelectQuery(BasicPattern.wrap(tAsList)), prototypicalGraph);
				ResultSet tripleQueryResult = tripleQueryExec.execSelect();
				
				// for each result from triple evaluation, do ...
				while(tripleQueryResult.hasNext()) {
					Binding tripleBinding = tripleQueryResult.nextBinding();
					
					// DEBUG
//					System.out.println("[[[\nTripleBinding -> " + tripleBinding + "\n]]]");
					
					BasicPattern appliedBGP = Substitute.substitute(reducedLocalTemplateBGP, tripleBinding);

					// DEBUG
//					System.out.println("[[[\n AppliedBGP -> " + appliedBGP + "\n]]]");

					// evaluate each applied BGP over the union of the prototypical graph and the TDB (aka available) graph
					QueryExecution appliedBGPExec = QueryExecutionFactory.create(QueryToolkit.bgpToSelectQuery(appliedBGP), model);
					ResultSet appliedBGPResult = appliedBGPExec.execSelect();
					
					// for each result from reduced BGP of completeness template, do ...
					while(appliedBGPResult.hasNext()) {

						// DEBUG
//						System.out.println("[[[\n AppliedBGPResult ->\n]]]");
						
						// compose and project the mapping
						BindingMap appliedBGPBinding = BindingFactory.create(appliedBGPResult.nextBinding());
						appliedBGPBinding.addAll(tripleBinding);
						BindingProject projectedAppliedBGPBinding = new BindingProject(localTemplateVars, appliedBGPBinding);
						
						// DEBUG
//						System.out.println("[[[\n Check Binding -> ]]]" + localTemplate.bindings + "\n Projected: " + BindingFactory.create(projectedAppliedBGPBinding));
//						System.out.println("Projected: " + BindingFactory.create(projectedAppliedBGPBinding));
						
						if(localTemplate.containsBinding(BindingFactory.create(projectedAppliedBGPBinding))) {

							// DEBUG
//							System.out.println("[[[\n Addingcruc ->\n]]]");
							
							crucPart.add(QueryToolkit.freeze(Substitute.substitute(BasicPattern.wrap(tAsList), tripleBinding)));
						}
					}
				}
				
			}
		}
		
		prototypicalGraph.remove(existingPrototypicalGraph);
		this.removeFromGraph(prototypicalGraph);
		return crucPart;
	}

	public Model getCrucialPartSuperOptimized(BasicPattern bgpToCheck, Set<CompletenessTemplate> templates) {
		
		boolean debug = false;
		
		// DEBUG	
		if(debug){
			System.out.println("##############################################################");
			System.out.println("### GET CRUCIAL PART FOR " + bgpToCheck);
			System.out.println("##############################################################");
		}
		
		Model prototypicalGraph = QueryToolkit.freeze(bgpToCheck);
		Model crucPart = ModelFactory.createDefaultModel();
		
		Model model = null;	
		model = ds.getDefaultModel();
		Iterator<CompletenessTemplate> templatesIt = templates.iterator();
		
		while(templatesIt.hasNext()) { // for each template, do ...
			
			CompletenessTemplate localTemplate = templatesIt.next();
			List<Var> localTemplateVars = localTemplate.getPseudoVars();
			BasicPattern localTemplateBGP = localTemplate.getBGP();
			
			// DEBUG	
			if(debug)
				System.out.println("LOCAL TEMPLATE BGP: " + localTemplateBGP);
			
			List<BasicPattern> nonEmptySubsetsOfLocalTemplateBGP = search(localTemplateBGP.getList());
			
			// DEBUG
			if(debug)
				System.out.println("SUBSETS: " + nonEmptySubsetsOfLocalTemplateBGP);
			
			// remove the empty set
			nonEmptySubsetsOfLocalTemplateBGP.remove(0);

			// DEBUG	
//			System.out.println("NON EMPTY SUBSETS: " + nonEmptySubsetsOfLocalTemplateBGP);
			
			// OPTMIZATION: we try to get all first
//			for(BasicPattern nonEmptySubset:nonEmptySubsetsOfLocalTemplateBGP) {
			for(int i = (nonEmptySubsetsOfLocalTemplateBGP.size()-1); i >= 0; i--) {
				
				BasicPattern nonEmptySubset = nonEmptySubsetsOfLocalTemplateBGP.get(i);
				
				QueryExecution nonEmptySubsetExec = QueryExecutionFactory.create(QueryToolkit.bgpToSelectQuery(nonEmptySubset), prototypicalGraph);
				ResultSet nonEmptySubsetQueryResult = nonEmptySubsetExec.execSelect();

				List<Triple> restOfNonEmptySubset = new ArrayList<Triple>(localTemplateBGP.getList());
				restOfNonEmptySubset.removeAll(nonEmptySubset.getList());
				BasicPattern restOfNonEmptySubsetAsBGP = BasicPattern.wrap(restOfNonEmptySubset);
				
				// DEBUG
				if(debug)
					System.out.println("REST BGP: " + restOfNonEmptySubsetAsBGP);
				
				while(nonEmptySubsetQueryResult.hasNext()) {
					
					Binding nonEmptySubsetQueryResultBinding = nonEmptySubsetQueryResult.nextBinding();
					// then apply to the rest
					BasicPattern appliedRestBGP = Substitute.substitute(restOfNonEmptySubsetAsBGP, nonEmptySubsetQueryResultBinding);
					
					// DEBUG
					if(debug)
						System.out.println("APPLIED REST BGP: " + appliedRestBGP);
					
					// if the rest contains a frozen mapping, then do not use this subset
					// otherwise, evaluate the rest over the data graph
					if(!appliedRestBGP.toString().contains("frozen")) {
						
						// evaluate each applied BGP over the union of the prototypical graph and the TDB (aka available) graph
						QueryExecution appliedRestBGPExec = QueryExecutionFactory.create(QueryToolkit.bgpToSelectQuery(appliedRestBGP), model);
						ResultSet appliedRestBGPResult = appliedRestBGPExec.execSelect();
						
						// for each result from reduced BGP of completeness template, do ...
						while(appliedRestBGPResult.hasNext()) {
							// compose and project the mapping
							BindingMap appliedRestBGPBinding = BindingFactory.create(appliedRestBGPResult.nextBinding());
							appliedRestBGPBinding.addAll(nonEmptySubsetQueryResultBinding);
							BindingProject projectedAppliedRestBGPBinding = new BindingProject(localTemplateVars, appliedRestBGPBinding);
							
							// DEBUG
							if(debug)
								System.out.println("WHOLE BINDING: " + appliedRestBGPBinding + ", Projected: " + projectedAppliedRestBGPBinding);
							
							// then for all the bindings of the rest, check if the projection of the bindings wrt meta variables exists in the template, if yes, return the frozen non empty subset
							
							if(localTemplate.containsBinding(BindingFactory.create(projectedAppliedRestBGPBinding))) {
								
								crucPart.add(QueryToolkit.freeze(Substitute.substitute(nonEmptySubset, nonEmptySubsetQueryResultBinding)));
								
								// OPTIMIZATION
								if(crucPart.size() == prototypicalGraph.size())
									return crucPart;
								
								// DEBUG
								if(debug)
									System.out.println("CRUC: " + crucPart);
								
								
							}
						}
					}
					
				}	
				
			}
		}
		return crucPart;
	}
	
	////////////////////////////////////////////////
	// method for generating schema-level completeness templates from queries
	////////////////////////////////////////////////

	public List<CompletenessTemplate> genCompletenessTemplatesForPatternSoundness(List<QueryWithNegation> queriesWithNegation) {
		
		List<CompletenessTemplate> completenessTemplates = new ArrayList<CompletenessTemplate>();
		for(QueryWithNegation queryWithNegation : queriesWithNegation) {
			completenessTemplates.addAll(genCompletenessTemplatesForPatternSoundness(queryWithNegation));
		}
		return completenessTemplates;

	}

	public List<CompletenessTemplate> genCompletenessTemplatesForPatternSoundness(QueryWithNegation queryWithNegation) {

		List<CompletenessTemplate> completenessTemplates = new ArrayList<CompletenessTemplate>();
		for(BasicPattern negPart : queryWithNegation.negParts) {
			completenessTemplates.add(genCompletenessTemplatesForPatternSoundness(queryWithNegation.posPart, negPart));
		}
		return completenessTemplates;
		
	}

	public CompletenessTemplate genCompletenessTemplatesForPatternSoundness(
			BasicPattern posPart, BasicPattern negPart) {
		
		List<Triple> patternList = new ArrayList<Triple>(posPart.getList());
		patternList.addAll(negPart.getList());
		Set<BindingMap> bindings = new HashSet<BindingMap>();		
		bindings.add(BindingFactory.create());
		return new CompletenessTemplate(patternList, bindings);
		
	}
	
	////////////////////////////////////////////////
	// (for 2 triple patterns) method for generating instance-level completeness templates from queries where each triple pattern of negative part is used
	////////////////////////////////////////////////

	public List<CompletenessTemplate> genCompletenessTemplatesForAnswerSoundnessPerTriplePattern(List<QueryWithNegation> queriesWithNegation) {
	
		List<CompletenessTemplate> completenessTemplates = new ArrayList<CompletenessTemplate>();
		for(QueryWithNegation queryWithNegation : queriesWithNegation) {
			completenessTemplates.addAll(genCompletenessTemplatesForAnswerSoundnessPerTriplePattern(queryWithNegation));
		}
		return completenessTemplates;
	
	}

	public List<CompletenessTemplate> genCompletenessTemplatesForAnswerSoundnessPerTriplePattern(QueryWithNegation queryWithNegation) {

		List<CompletenessTemplate> completenessTemplates = new ArrayList<CompletenessTemplate>();
		List<Binding> bindings = this.evalSELECT(queryWithNegation.getQuery());
	
		// MAX BINDINGS
//		if(bindings.size() > 10)
//			bindings = bindings.subList(0, 10);
		
		maxBindings = new Double(Math.ceil(percentageOfBindings*bindings.size())).intValue();
		bindings = bindings.subList(0, maxBindings);
		
		// DEBUG
		System.out.println("Generating templates with " + bindings.size() + " bindings.\n---");
		
		if(!bindings.isEmpty()) {
			for(BasicPattern negPart : queryWithNegation.negParts) {
				completenessTemplates.addAll(genCompletenessTemplatesForAnswerSoundnessPerTriplePattern(bindings, negPart));
			}
		}
		return completenessTemplates;

	}
	
	public List<CompletenessTemplate> genCompletenessTemplatesForAnswerSoundnessPerTriplePattern(List<Binding> bindings, BasicPattern negPart) {

		List<CompletenessTemplate> completenessTemplates = new ArrayList<CompletenessTemplate>();
		List<Triple> triplePatterns = negPart.getList();
		List<Var> posVars = QueryToolkit.getVarsFromBindings(bindings);
		List<Var> negVars = QueryToolkit.getVars(triplePatterns);
		List<Var> intersectionVarsWithAllNETriplePatterns = intersectionVar(posVars, negVars);
		
		Set<BindingMap> intersectedBindingsWithAllNETriplePatterns = new HashSet<BindingMap>();
		
		for(Binding binding : bindings) {
				BindingProject projectedBindingToIntersectionVars = new BindingProject(intersectionVarsWithAllNETriplePatterns, binding);
				intersectedBindingsWithAllNETriplePatterns.add(BindingFactory.create(BindingFactory.materialize(projectedBindingToIntersectionVars))); // materialization is needed to create a new copy of the binding, otherwise it will be mapped to NULL
		}
		
		// for first triple pattern
		List<Triple> firstTriplePatternAsList = triplePatterns.subList(0, 1);
		List<Var> negVarsWithFirstTriplePattern = QueryToolkit.getVars(firstTriplePatternAsList);
		List<Var> intersectionVarsWithFirstTriplePattern = intersectionVar(posVars, negVarsWithFirstTriplePattern);
		
		Set<BindingMap> intersectedBindingsWithFirstTriplePattern = new HashSet<BindingMap>();
		Set<BindingMap> intersectedBindingsWithAllNETriplePatternsPlusFirstTriplePattern = new HashSet<BindingMap>();
		
		// debug
//		System.out.println("intersectedBindingsWithAllNETriplePatterns: " + intersectedBindingsWithAllNETriplePatterns.size());
//		System.out.println("intersectionVarsWithFirstTriplePattern: " + intersectionVarsWithFirstTriplePattern);
		
		Iterator<BindingMap> intersectedBindingsWithAllNETriplePatternsIt = intersectedBindingsWithAllNETriplePatterns.iterator();
		while(intersectedBindingsWithAllNETriplePatternsIt.hasNext()) {
			
			BindingMap binding = intersectedBindingsWithAllNETriplePatternsIt.next();
			BindingProject projectedBindingToIntersectionVars = new BindingProject(intersectionVarsWithFirstTriplePattern, binding);
			intersectedBindingsWithFirstTriplePattern.add(BindingFactory.create(BindingFactory.materialize(projectedBindingToIntersectionVars))); // materialization is needed to create a new copy of the binding, otherwise it will be mapped to NULL
			BasicPattern instantiatedFirstTriplePattern = Substitute.substitute(BasicPattern.wrap(firstTriplePatternAsList), projectedBindingToIntersectionVars);
			List<Binding> bindingsFromFirstTriplePattern = this.evalSELECT(QueryToolkit.bgpToSelectQuery(instantiatedFirstTriplePattern));
			
			// debug
//			System.out.println("instantiatedFirstTriplePattern: " + instantiatedFirstTriplePattern);
//			System.out.println("bindingsFromFirstTriplePattern: " + bindingsFromFirstTriplePattern);
			
			for(Binding bindingFromFirstTriplePattern:bindingsFromFirstTriplePattern) {
				Binding mergedBinding = BindingFactory.materialize(BindingUtils.merge(binding, bindingFromFirstTriplePattern));
				
				// debug
//				System.out.println("HERE: " + mergedBinding);
				
				intersectedBindingsWithAllNETriplePatternsPlusFirstTriplePattern.add(BindingFactory.create(mergedBinding));
			}
		}
		
		// debug
//		System.out.println("intersectedBindingsWithAllNETriplePatternsPlusFirstTriplePattern: " + intersectedBindingsWithAllNETriplePatternsPlusFirstTriplePattern.size());
		
		completenessTemplates.add(new CompletenessTemplate(firstTriplePatternAsList, intersectedBindingsWithFirstTriplePattern));
		
		// for second triple pattern
		List<Triple> secondTriplePatternAsList = triplePatterns.subList(1, 2);
		List<Var> negVarsWithSecondTriplePattern = QueryToolkit.getVars(secondTriplePatternAsList);
		Set<Var> unionVars = new HashSet<Var>();
		unionVars.addAll(posVars);
		unionVars.addAll(negVarsWithFirstTriplePattern);
		List<Var> intersectionVarsWithSecondTriplePattern = intersectionVar(new ArrayList<Var>(unionVars), negVarsWithSecondTriplePattern);
		Set<BindingMap> intersectedBindingsWithSecondTriplePattern = new HashSet<BindingMap>();
		
		//DEBUG
//		System.out.println(secondTriplePatternAsList);
//		System.out.println(negVarsWithSecondTriplePattern);
//		System.out.println(intersectionVarsWithSecondTriplePattern);
		
		Iterator<BindingMap> intersectedBindingsWithAllNETriplePatternsPlusFirstTriplePatternIt = intersectedBindingsWithAllNETriplePatternsPlusFirstTriplePattern.iterator();
		while(intersectedBindingsWithAllNETriplePatternsPlusFirstTriplePatternIt.hasNext()) {
			BindingMap binding = intersectedBindingsWithAllNETriplePatternsPlusFirstTriplePatternIt.next();
			BindingProject projectedBindingToIntersectionVars = new BindingProject(intersectionVarsWithSecondTriplePattern, binding);
			intersectedBindingsWithSecondTriplePattern.add(BindingFactory.create(BindingFactory.materialize(projectedBindingToIntersectionVars)));
		}
		
		if(!intersectedBindingsWithSecondTriplePattern.isEmpty())
			completenessTemplates.add(new CompletenessTemplate(secondTriplePatternAsList, intersectedBindingsWithSecondTriplePattern));
		
		return completenessTemplates;

	}
	
	////////////////////////////////////////////////
	// method for generating instance-level completeness templates from queries where full body of negative part is used
	////////////////////////////////////////////////
	
	// idea: evaluate whole query, generate templates from the (some percentage of) bindings applied to the negative part
	public List<CompletenessTemplate> genCompletenessTemplatesForAnswerSoundness(List<QueryWithNegation> queriesWithNegation) {
		
		List<CompletenessTemplate> completenessTemplates = new ArrayList<CompletenessTemplate>();
		for(QueryWithNegation queryWithNegation : queriesWithNegation) {
			completenessTemplates.addAll(genCompletenessTemplatesForAnswerSoundness(queryWithNegation));
		}
		return completenessTemplates;
		
	}

	public List<CompletenessTemplate> genCompletenessTemplatesForAnswerSoundness(
			QueryWithNegation queryWithNegation) {

		List<CompletenessTemplate> completenessTemplates = new ArrayList<CompletenessTemplate>();
		List<Binding> bindings = this.evalSELECT(queryWithNegation.getQuery());
		
		if(!bindings.isEmpty()) {
			
			for(BasicPattern negPart : queryWithNegation.negParts) {
				completenessTemplates.add(genCompletenessTemplatesForAnswerSoundness(bindings, negPart));
			}
		}
		return completenessTemplates;
	
	}
	
	// auxiliary method
	public static List<Var> intersectionVar(List<Var> vars1, List<Var> vars2) {		
		
		List<Var> intersectionList = new ArrayList<Var>();
		for(Var v : vars1) {
			if(vars2.contains(v))
				intersectionList.add(v);
		}
		
		return intersectionList;
	}
	
	int maxBindings = 10000;
	double percentageOfBindings = 0.25;
	
	public CompletenessTemplate genCompletenessTemplatesForAnswerSoundness(
			List<Binding> bindings, BasicPattern negPart) {

		// CONFIGURATION: comment this to set max bindings to 10000
		maxBindings = new Double(Math.ceil(percentageOfBindings*bindings.size())).intValue();
		
		// DEBUG
		System.out.println("Generating templates with " + maxBindings + " bindings.\n---");
		
		List<Var> posVars = QueryToolkit.getVarsFromBindings(bindings);
		List<Var> negVars = QueryToolkit.getVars(negPart.getList());
		List<Var> intersectionVars = intersectionVar(posVars, negVars);
		
		Set<BindingMap> intersectedBindings = new HashSet<BindingMap>();

		int currentBindingCounter = 0;
		for(Binding binding : bindings) {
			if(currentBindingCounter < maxBindings) {
				BindingProject projectedBindingToIntersectionVars = new BindingProject(intersectionVars, binding);
//				intersectedBindings.add(BindingFactory.create(projectedBindingToIntersectionVars));
				intersectedBindings.add(BindingFactory.create(BindingFactory.materialize(projectedBindingToIntersectionVars))); // materialization is needed to create a new copy of the binding, otherwise it will be mapped to NULL
			}
			currentBindingCounter++;
		}

		return new CompletenessTemplate(negPart.getList(), intersectedBindings);
		
	}

	// completeness checking with the partial match technique only!
	public boolean checkCompletenessPMatchOnly(BasicPattern bgp,
			HashMap<String, List<CompletenessTemplate>> templateMap) {
		
		try {
			saturateAndCheckContainmentPMatchOnly(bgp, templateMap, new HashSet<String>());
		}
		catch(IncompletenessException ie) {
			return false;
		}
		return true;
		
	}

	private void saturateAndCheckContainmentPMatchOnly(BasicPattern bgp,
			HashMap<String, List<CompletenessTemplate>> templateMap, Set<String> checkedSignatures) throws IncompletenessException {

		// DEBUG
//		System.out.println("BEFORE EPG OPTIMIZED");
		
		Set<BasicPattern> Pequiv = epgPMatchOnly(bgp, templateMap, checkedSignatures); // we need not check again the parts that are known useless for completeness checking
		
		// DEBUG
//		System.out.println("AFTER EPG OPTIMIZED");
//		System.out.println(bgp + " ### " + Pequiv);
		
		// base case
		if(Pequiv.size() != 0 && Pequiv.iterator().next().toString().compareTo(bgp.toString()) == 0) {
			if(!(this.contains(QueryToolkit.freeze(bgp))))
				throw new IncompletenessException();
		}
		else if(Pequiv.size() != 0) {
			
			Set<String> newCheckedSignatures = SignatureFactory.getSignatureWithGeneralization(intersection(bgp, Pequiv.iterator().next()));
			Iterator<BasicPattern> PequivIt = Pequiv.iterator();
			while(PequivIt.hasNext()) {
				BasicPattern thisBGP = PequivIt.next();
				saturateAndCheckContainmentPMatchOnly(thisBGP, templateMap, newCheckedSignatures);
			}
		}
	}

	private Set<BasicPattern> epgPMatchOnly(BasicPattern bgp,
			HashMap<String, List<CompletenessTemplate>> templateMap,
			Set<String> checkedSignatures) {
	
		// DEBUG
//		System.out.println("[[[\nTempMap -> " + templateMap.values().iterator().next().get(0).bindings + "\n]]]");
		
		Set<CompletenessTemplate> partiallyMatchedTemplates = partialMatch(bgp, checkedSignatures, templateMap);
		
		Set<BasicPattern> bgps = new HashSet<BasicPattern>();
		
		List<CompletenessStatement> partiallyMatchedTemplatesStatements = new ArrayList<CompletenessStatement>();
		for(CompletenessTemplate partiallyMatchedTemplate:partiallyMatchedTemplates) {
			partiallyMatchedTemplatesStatements.addAll(partiallyMatchedTemplate.unfold());
		}
		
		// DEBUG
//		System.out.println("Evaluating..." + partiallyMatchedTemplatesStatements.size() + " pmatch statements.");
//		long startCrucPart = System.nanoTime();
		BasicPattern crucPart = this.getCrucialPartNaive(bgp, partiallyMatchedTemplatesStatements);
//		System.out.println("Cruc part pmatch takes (ns): " + (System.nanoTime()-startCrucPart));

		// optimization rule: completeness skip
		// NOTE: this optimization rule assumes that BGPs contain no duplicates
		if(crucPart.size() == bgp.size()) {
			return bgps;
		}
		
		if(crucPart.size() == 0) {
			bgps.add(bgp);
			return bgps;
		}
		
		List<Triple> bgpToInstantiate = new ArrayList<Triple>(bgp.getList());
		
		// optimization rule: BGP length reduction
		bgpToInstantiate.removeAll(crucPart.getList());
		
		// optimization rule: branching reduction
		bgps = this.instantiateNaive(crucPart, BasicPattern.wrap(bgpToInstantiate));
		
		return bgps;
		
	}

	// methods for generating all subsets of a BGP
	private List<BasicPattern> nonEmptySubsetsOfBGPs;
	
	public List<BasicPattern> search(List<Triple> BGP) {
		nonEmptySubsetsOfBGPs = new ArrayList<BasicPattern>(); // reset the content of relevantCSList
		recSearch(new ArrayList<Triple>(), BGP, 0);
		return nonEmptySubsetsOfBGPs;
	}

	// credit to Werner Nutt: idea is flipping bits between 0 and 1
	private void recSearch(List<Triple> prefix,
			List<Triple> predicateListOfQuery, int level) {
		
		if(level > predicateListOfQuery.size() - 1) {
			nonEmptySubsetsOfBGPs.add(BasicPattern.wrap(prefix));
		}
		else {
			List<Triple> nonGrowingPrefix = new ArrayList<Triple>();
			List<Triple> growingPrefix = new ArrayList<Triple>();
			nonGrowingPrefix.addAll(prefix);
			growingPrefix.addAll(prefix);
			growingPrefix.add(predicateListOfQuery.get(level));
			recSearch(nonGrowingPrefix, predicateListOfQuery, level + 1);
			recSearch(growingPrefix, predicateListOfQuery, level + 1);
		}
		
	}
	
}
