package it.unibz.inf.completeness;

import it.unibz.inf.semweb.QueryToolkit;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import com.hp.hpl.jena.graph.Node;
import com.hp.hpl.jena.graph.Triple;
import com.hp.hpl.jena.sparql.core.BasicPattern;
import com.hp.hpl.jena.sparql.core.Substitute;
import com.hp.hpl.jena.sparql.core.Var;
import com.hp.hpl.jena.sparql.engine.binding.Binding;
import com.hp.hpl.jena.sparql.engine.binding.BindingFactory;
import com.hp.hpl.jena.sparql.engine.binding.BindingMap;

// a representation (or data structure) of completeness statements of the same form
public class CompletenessTemplate {

	public CompletenessStatement cs;
	public Set<BindingMap> bindings; // benefits of using BindingMap: Substitute, addAll (for binding composition), BindingProject (for variable projection)

	// constructors
	public CompletenessTemplate(BasicPattern bgp, Set<BindingMap> bindings) {
		this.cs = new CompletenessStatement(bgp);
		this.bindings = new HashSet<BindingMap>(bindings);
	}
	
	public CompletenessTemplate(List<Triple> bgp, Set<BindingMap> bindings) {
		this.cs = new CompletenessStatement(bgp);
		this.bindings = new HashSet<BindingMap>(bindings);
	}

	// methods
	public BasicPattern getBGP() {
		return cs.getBGP();
	}

	public boolean containsBinding(BindingMap binding) {
		return bindings.contains(binding);
	}
	
	public List<CompletenessStatement> unfold() {
		List<CompletenessStatement> csList = new ArrayList<CompletenessStatement>();
		BasicPattern csBGP = this.getBGP();
		for(BindingMap binding : bindings) {
			csList.add(new CompletenessStatement(Substitute.substitute(csBGP, binding)));
		}
		return csList;
	}

	public List<Var> getPseudoVars() {
		List<Var> vars = new ArrayList<Var>();
		if(bindings.iterator().hasNext()) {
			Iterator<Var> varIt = bindings.iterator().next().vars(); // get the variables of one mapping
			while(varIt.hasNext())
				vars.add(varIt.next());
		}
		// NOTE: A template must have at least one mapping (it can also be an empty mapping), so the following printing must not occur!
		else {
			System.out.println("ERROR: " + this.toString() + " : " + this.bindings);
			System.exit(0);
		}
		return vars;
	}
	
	public List<String> getPseudoVarsAsString() {
		List<Var> vars = getPseudoVars();
		List<String> varsAsString = new ArrayList<String>();
		for(Var var : vars)
			varsAsString.add(var.toString());
		Collections.sort(varsAsString);
		return varsAsString;
	}
	
	public Set<String> getSignature() {
		return cs.getSignature();
	}

	// NOTE: this has a drawback if there's variable name playing
	public String toString() {
		return cs.toString() + getPseudoVarsAsString();
	}

	// to calculate hash key
	public int hashCode(){
	    return this.toString().hashCode();
	}
	
	public boolean equals(Object o){
	    if(o == null) return false;
	    
	    if(!(o instanceof CompletenessTemplate)) return false;

	    CompletenessTemplate other = (CompletenessTemplate) o;
	    if(this.toString().compareTo(other.toString()) != 0) return false;

	    return true;
	}
	
	//////////////////////////////////////
	// Removing Redundancies Methods
	//////////////////////////////////////
	
	public static List<CompletenessTemplate> removeRedundancies(List<CompletenessTemplate> oldCompletenessTemplates) {
		
		if(oldCompletenessTemplates.size() == 0 || oldCompletenessTemplates.size() == 1) {
			return oldCompletenessTemplates;
		}
		
		List<CompletenessTemplate> newCompletenessTemplates = new ArrayList<CompletenessTemplate>(oldCompletenessTemplates);
		
		int size = 0;
		while(newCompletenessTemplates.size() != size) {
			size = newCompletenessTemplates.size();
			newCompletenessTemplates = removeRedundanciesOneLayer(newCompletenessTemplates);
		}
		
		return newCompletenessTemplates;
		
	}
	
	public static List<CompletenessTemplate> removeRedundanciesOneLayer(
			List<CompletenessTemplate> completenessTemplates) {
		
		if(completenessTemplates.size() == 0 || completenessTemplates.size() == 1) {
			return completenessTemplates;
		}
		
		List<CompletenessTemplate> newTemplates = new ArrayList<CompletenessTemplate>();
		
		while(completenessTemplates.size() != 0) {
			CompletenessTemplate currentTemplate = completenessTemplates.get(0);
			completenessTemplates.remove(0);
			
			for (Iterator<CompletenessTemplate> iterator = completenessTemplates.iterator(); iterator.hasNext();) {
				CompletenessTemplate thisTemplate = iterator.next();
				CompletenessTemplate unifiedTemplates = unifyTemplates(currentTemplate, thisTemplate);
				if (unifiedTemplates != null) {
					currentTemplate = unifiedTemplates;
					// Remove the current element from the iterator and the list.
					iterator.remove();
				}
			}
			newTemplates.add(currentTemplate);
		}
			
		return newTemplates;
	}

	// can the second template be unified with the first template?
	public static CompletenessTemplate unifyTemplates(CompletenessTemplate template1, CompletenessTemplate template2) {
		
		List<Binding> bindings = QueryToolkit.areBGPsEquivalent(template1.getBGP(), template2.getBGP());
		
		if(bindings.isEmpty()) return null;
		
		List<Var> template1Vars = template1.getPseudoVars();
		List<Var> template2Vars = template2.getPseudoVars();

		// DEBUG
//		System.out.println(bindings);
		
		HashMap<Var, Var> varHashmap1 = null;
		for(Binding binding : bindings) {
			varHashmap1 = containVars(template1Vars, template2Vars, binding);
			HashMap<Var, Var> varHashmap2 = containVars(template2Vars, template1Vars, inverse(binding));
			
			// DEBUG
//			System.out.println(varHashmap1);
//			System.out.println(varHashmap2);
			
			if(varHashmap1 != null && varHashmap2 != null){
				
				// DEBUG
//				System.out.println(varHashmap1);
				
				break;
			}
		}
		
		if(varHashmap1 == null)
			return null;
		
		Set<BindingMap> newBindingMapForTemplate2 = translateMappings(template2.bindings, varHashmap1);
		newBindingMapForTemplate2.addAll(template1.bindings);
		
		CompletenessTemplate unifiedTemplate = new CompletenessTemplate(template1.getBGP(), newBindingMapForTemplate2);
	
		// DEBUG
//		System.out.println("Content: " + unifiedTemplate.bindings);
//		System.out.println("Size: " + unifiedTemplate.bindings.size());

		return unifiedTemplate;
		
	}

	public static Set<BindingMap> translateMappings(Set<BindingMap> bindings2,
			HashMap<Var, Var> varHashmap1) {

		Set<BindingMap> bindings = new HashSet<BindingMap>();
		Iterator<BindingMap> bindings2It = bindings2.iterator();
		while(bindings2It.hasNext()) {
			BindingMap binding2 = bindings2It.next();
			bindings.add(apply(binding2, varHashmap1));
		}
		
		return bindings;
	}

	public static BindingMap apply(BindingMap binding2,
			HashMap<Var, Var> varHashmap1) {
		
		BindingMap newBinding = BindingFactory.create();
		Iterator<Var> varIt = binding2.vars();
		while(varIt.hasNext()) {
			Var oldVar = varIt.next();
			Var newVar = varHashmap1.get(oldVar);
			newBinding.add(newVar, binding2.get(oldVar));
		}
		
		return newBinding;
	}

	public static Binding inverse(Binding binding) {
		
		BindingMap inverseBinding = BindingFactory.create();
		Iterator<Var> varIt = binding.vars();
		while(varIt.hasNext()) {
			Var thisVar = varIt.next();
			Var newVar = Var.alloc(QueryToolkit.defreeze(binding.get(thisVar)));
			Node newNode = QueryToolkit.freeze(thisVar);
			inverseBinding.add(newVar, newNode);
		}
		return inverseBinding;
		
	}

	public static HashMap<Var, Var> containVars(List<Var> template1Vars,
			List<Var> template2Vars, Binding binding) {
		
		HashMap<Var, Var> varHashmap = new HashMap<Var, Var>();
		for(Var template2Var : template2Vars) {
			Var newVar = applyBinding(binding, template2Var);
			if(!template1Vars.contains(newVar))
				return null;
			else
				varHashmap.put(template2Var, newVar);
		}
		return varHashmap;
		
	}

	public static Var applyBinding(Binding binding, Var template2Var) {
		
		Node node = binding.get(template2Var);
		if(node == null)
			return null;
		else
			return Var.alloc(QueryToolkit.defreeze(node));
			
	}
	
}
