package it.unibz.inf.completeness;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import com.hp.hpl.jena.graph.Node;
import com.hp.hpl.jena.graph.Triple;
import com.hp.hpl.jena.sparql.core.BasicPattern;

public class SignatureFactory {

	//////////////////////////////////////
	// signature generation from various types
	//////////////////////////////////////
	
	// generate the signature map of completeness templates
	// for faster mapping, strings (instead of triples) are used as keys
	public static HashMap<String, List<CompletenessTemplate>> getSignatureMapForTemplates(List<CompletenessTemplate> csTemplates) {
		HashMap<String, List<CompletenessTemplate>> signatureMap = new HashMap<String, List<CompletenessTemplate>>();
		
		for(CompletenessTemplate csTemplate : csTemplates) {
			
			// DEBUG
//			System.out.println("GETSIGNATURE: " + csTemplate.getSignature());
//			System.out.println("BINDINGS: " + csTemplate.bindings);
			
			for(String signatureTriple : csTemplate.getSignature()) {
				if(signatureMap.containsKey(signatureTriple)) {
					signatureMap.get(signatureTriple).add(csTemplate);
				}
				else {
					List<CompletenessTemplate> csTemplatesInitial = new ArrayList<CompletenessTemplate>();
					csTemplatesInitial.add(csTemplate);
					signatureMap.put(signatureTriple, csTemplatesInitial);
				}
			}
		}
		
		return signatureMap;
	}

	public static Set<String> getSignature(CompletenessStatement cs) {
		return getSignature(cs.getBGP());
	}

	public static Set<String> getSignature(BasicPattern bgp) {
		Set<String> bgpSignature = new HashSet<String>();
		for(Triple t : bgp) {
			bgpSignature.add(getSignature(t));
		}
		return bgpSignature;
	}
	
	public static Set<String> getSignature(List<Triple> bgp) {
		Set<String> bgpSignature = new HashSet<String>();
		for(Triple t : bgp) {
			bgpSignature.add(getSignature(t));
		}
		return bgpSignature;
	}

	// example signature: http://tutorial-academy.com/2015/tdb#John->http://tutorial-academy.com/2015/tdb#owns->?VAR
	// the signature created must be consistent with the signature generalization method below
	public static String getSignature(Triple t) {
		return getSignature(t.getSubject()) + "->" + getSignature(t.getPredicate()) + "->" + getSignature(t.getObject());
	}

	public static String getSignature(Node n) {
		if(n.isVariable())
			return "?VAR";
		else
			return n.toString();
	}
	
	//////////////////////////////////////
	// partial matching
	//////////////////////////////////////

	public static Set<CompletenessTemplate> getCompletenessTemplatesFromSignatures(Set<String> signatures, HashMap<String, List<CompletenessTemplate>> signatureMap) {
		Set<CompletenessTemplate> templates = new HashSet<CompletenessTemplate>();
		for(String signature : signatures) {
			List<CompletenessTemplate> templateWithThisSignature = signatureMap.get(signature);
			if(templateWithThisSignature != null)
				templates.addAll(templateWithThisSignature);
		}
		return templates;
	}
	
	// get unchecked signatures of the signature generalization of a BGP
	public static Set<String> getUncheckedSignatures(List<Triple> bgp, Set<String> checkedSignatures) {
		Set<String> uncheckedSignatures = getSignatureWithGeneralization(bgp);
		uncheckedSignatures.removeAll(checkedSignatures);
		return uncheckedSignatures;
	}
	
	public static Set<String> getSignatureWithGeneralization(List<Triple> bgp) {
		Set<String> generalizedSignatureBGP = new HashSet<String>();
		for(Triple t : bgp) {
			generalizedSignatureBGP.addAll(getSignatureWithGeneralization(t));
		}
		return generalizedSignatureBGP;
	}

	// we do not generalize predicates as it is not likely that we can be complete for ALL predicates of things
	public static Set<String> getSignatureWithGeneralization(Triple t) {
		Set<String> generalizedSignatureTriple = new HashSet<String>();
		
		generalizedSignatureTriple.add(getSignature(t.getSubject()) + "->" + getSignature(t.getPredicate()) + "->" + getSignature(t.getObject()));
		generalizedSignatureTriple.add("?VAR->" + getSignature(t.getPredicate()) + "->" + getSignature(t.getObject()));
		generalizedSignatureTriple.add(getSignature(t.getSubject()) + "->" + getSignature(t.getPredicate()) + "->?VAR");
		generalizedSignatureTriple.add("?VAR->" + getSignature(t.getPredicate()) + "->?VAR");
		
		return generalizedSignatureTriple;
	}
	
}
