package it.unibz.inf.semweb;

import java.io.IOException;
import java.nio.charset.Charset;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import com.hp.hpl.jena.graph.Node;
import com.hp.hpl.jena.graph.NodeFactory;
import com.hp.hpl.jena.graph.Triple;
import com.hp.hpl.jena.query.Query;
import com.hp.hpl.jena.query.QueryExecution;
import com.hp.hpl.jena.query.QueryExecutionFactory;
import com.hp.hpl.jena.query.QueryFactory;
import com.hp.hpl.jena.query.ResultSet;
import com.hp.hpl.jena.rdf.model.Model;
import com.hp.hpl.jena.rdf.model.ModelFactory;
import com.hp.hpl.jena.rdf.model.ResourceFactory;
import com.hp.hpl.jena.rdf.model.Statement;
import com.hp.hpl.jena.rdf.model.Property;
import com.hp.hpl.jena.rdf.model.RDFNode;
import com.hp.hpl.jena.rdf.model.Resource;
import com.hp.hpl.jena.rdf.model.StmtIterator;
import com.hp.hpl.jena.sparql.core.BasicPattern;
import com.hp.hpl.jena.sparql.core.PathBlock;
import com.hp.hpl.jena.sparql.core.Var;
import com.hp.hpl.jena.sparql.syntax.Element;
import com.hp.hpl.jena.sparql.syntax.ElementFilter;
import com.hp.hpl.jena.sparql.syntax.ElementGroup;
import com.hp.hpl.jena.sparql.syntax.ElementPathBlock;
import com.hp.hpl.jena.sparql.syntax.ElementTriplesBlock;
import com.hp.hpl.jena.sparql.syntax.Template;
import com.hp.hpl.jena.sparql.core.TriplePath;
import com.hp.hpl.jena.sparql.engine.binding.Binding;
import com.hp.hpl.jena.sparql.expr.E_NotExists;

// methods for SPARQL query manipulation
public class QueryToolkit {

	//////////////////////////////////////
	// Variable Extraction
	//////////////////////////////////////
	
	// get variables from the first binding of bindings
	public static List<Var> getVarsFromBindings(List<Binding> bindings) {
		
		List<Var> vars = new ArrayList<Var>();
		if(bindings.iterator().hasNext()) {
			Iterator<Var> varIt = bindings.iterator().next().vars(); // get the variables
			while(varIt.hasNext())
				vars.add(varIt.next());
		}
		return vars;
		
	}
	
	// get (a set of) variables from a BGP
	public static List<Var> getVars(List<Triple> triples) {
		
		Set<Var> vars = new HashSet<Var>();
		for(Triple triple : triples) {
			vars.addAll(getVars(triple));
		}
		return new ArrayList<Var>(vars);
		
	}

	// get variables from a triple pattern
	private static List<Var> getVars(Triple triple) {
		
		List<Var> vars = new ArrayList<Var>();
		if(triple.getSubject().isVariable())
			vars.add(Var.alloc(triple.getSubject()));
		if(triple.getPredicate().isVariable())
			vars.add(Var.alloc(triple.getPredicate()));
		if(triple.getObject().isVariable())
			vars.add(Var.alloc(triple.getObject()));
		return vars;
		
	}

	//////////////////////////////////////
	// Experiment Queries Generation
	//////////////////////////////////////

	// case oneTP: last triple pattern is negated
	public static List<QueryWithNegation> genQueriesWithNegationOneTP(List<Query> queries) {
		
		List<QueryWithNegation> queriesWithNegation = new ArrayList<QueryWithNegation>();
		for(Query query : queries) {
			queriesWithNegation.addAll(genQueriesWithNegationOneTP(query));
		}
		return queriesWithNegation;
		
	}
	
	public static List<QueryWithNegation> genQueriesWithNegationOneTP(Query query) {

		List<Triple> triples = QueryToolkit.body(query).getList();
		int triplesSize = triples.size();
		List<QueryWithNegation> queriesWithNegation = new ArrayList<QueryWithNegation>();
		if(triplesSize == 1) {
			BasicPattern posPart = new BasicPattern();
			BasicPattern negBGP = BasicPattern.wrap(triples);
			List<BasicPattern> negParts = new ArrayList<BasicPattern>();
			negParts.add(negBGP);
			queriesWithNegation.add(new QueryWithNegation(posPart, negParts));
		}
		else {
			BasicPattern posPart = BasicPattern.wrap(triples.subList(0, triplesSize-1));
			BasicPattern negBGP = BasicPattern.wrap(triples.subList(triplesSize-1, triplesSize));
			List<BasicPattern> negParts = new ArrayList<BasicPattern>();
			negParts.add(negBGP);
			queriesWithNegation.add(new QueryWithNegation(posPart, negParts));
		}
		return queriesWithNegation;
		
	}	
	
	// case oneTPoneTP: last two triple patterns are negated SEPARATELY
	public static List<QueryWithNegation> genQueriesWithNegationOneTPOneTP(List<Query> queries) {

		List<QueryWithNegation> queriesWithNegation = new ArrayList<QueryWithNegation>();
		for(Query query : queries) {
			queriesWithNegation.addAll(genQueriesWithNegationOneTPOneTP(query));
		}
		return queriesWithNegation;
		
	}
	
	public static List<QueryWithNegation> genQueriesWithNegationOneTPOneTP(Query query) {

		List<Triple> triples = QueryToolkit.body(query).getList();
		int triplesSize = triples.size();
		
		List<QueryWithNegation> queriesWithNegation = new ArrayList<QueryWithNegation>();
		if(triplesSize < 2) return queriesWithNegation; // we do not consider queries of length < 2
		else if(triplesSize == 2) {
			BasicPattern posPart = new BasicPattern();
			BasicPattern negBGP1 = BasicPattern.wrap(triples.subList(0, 1));
			BasicPattern negBGP2 = BasicPattern.wrap(triples.subList(1, 2));
			List<BasicPattern> negParts = new ArrayList<BasicPattern>();
			negParts.add(negBGP1);
			negParts.add(negBGP2);
			queriesWithNegation.add(new QueryWithNegation(posPart, negParts));
		}
		else {
			BasicPattern posPart = BasicPattern.wrap(triples.subList(0, triplesSize-2));
			BasicPattern negBGP1 = BasicPattern.wrap(triples.subList(triplesSize-2, triplesSize-1));
			BasicPattern negBGP2 = BasicPattern.wrap(triples.subList(triplesSize-1, triplesSize));
			List<BasicPattern> negParts = new ArrayList<BasicPattern>();
			negParts.add(negBGP1);
			negParts.add(negBGP2);
			queriesWithNegation.add(new QueryWithNegation(posPart, negParts));
		}
		
		return queriesWithNegation;
		
	}

	// case twoTPs: last two triple patterns are negated TOGETHER
	public static List<QueryWithNegation> genQueriesWithNegationTwoTPs(List<Query> queries) {

		List<QueryWithNegation> queriesWithNegation = new ArrayList<QueryWithNegation>();
		for(Query query : queries) {
			queriesWithNegation.addAll(genQueriesWithNegationTwoTPs(query));
		}
		return queriesWithNegation;
		
	}
	
	public static List<QueryWithNegation> genQueriesWithNegationTwoTPs(Query query) {

		List<Triple> triples = QueryToolkit.body(query).getList();
		int triplesSize = triples.size();
		
		List<QueryWithNegation> queriesWithNegation = new ArrayList<QueryWithNegation>();
		if(triplesSize < 2) return queriesWithNegation;
		else if(triplesSize == 2) {
			BasicPattern posPart = new BasicPattern();
			BasicPattern negBGP = BasicPattern.wrap(triples.subList(0, 2));
			List<BasicPattern> negParts = new ArrayList<BasicPattern>();
			negParts.add(negBGP);
			queriesWithNegation.add(new QueryWithNegation(posPart, negParts));
		}
		else {
			BasicPattern posPart = BasicPattern.wrap(triples.subList(0, triplesSize-2));
			BasicPattern negBGP = BasicPattern.wrap(triples.subList(triplesSize-2, triplesSize));
			List<BasicPattern> negParts = new ArrayList<BasicPattern>();
			negParts.add(negBGP);
			queriesWithNegation.add(new QueryWithNegation(posPart, negParts));
		}
		
		return queriesWithNegation;
		
	}

	// case threeTPs: last three triple patterns are negated TOGETHER
	public static List<QueryWithNegation> genQueriesWithNegationThreeTPs(List<Query> queries) {

		List<QueryWithNegation> queriesWithNegation = new ArrayList<QueryWithNegation>();
		for(Query query : queries) {
			queriesWithNegation.addAll(genQueriesWithNegationThreeTPs(query));
		}
		return queriesWithNegation;
		
	}

	public static List<QueryWithNegation> genQueriesWithNegationThreeTPs(Query query) {

		List<Triple> triples = QueryToolkit.body(query).getList();
		int triplesSize = triples.size();
		
		List<QueryWithNegation> queriesWithNegation = new ArrayList<QueryWithNegation>();
		if(triplesSize < 3) return queriesWithNegation;
		else if(triplesSize == 3) {
			BasicPattern posPart = new BasicPattern();
			BasicPattern negBGP = BasicPattern.wrap(triples.subList(0, 3));
			List<BasicPattern> negParts = new ArrayList<BasicPattern>();
			negParts.add(negBGP);
			queriesWithNegation.add(new QueryWithNegation(posPart, negParts));
		}
		else {
			BasicPattern posPart = BasicPattern.wrap(triples.subList(0, triplesSize-3));
			BasicPattern negBGP = BasicPattern.wrap(triples.subList(triplesSize-3, triplesSize));
			List<BasicPattern> negParts = new ArrayList<BasicPattern>();
			negParts.add(negBGP);
			queriesWithNegation.add(new QueryWithNegation(posPart, negParts));
		}
		
		return queriesWithNegation;
		
	}
	
	// generic method for generating a query with negation
	public static Query genQueryWithNegation(BasicPattern posPart, List<BasicPattern> negParts) {
		
		ElementTriplesBlock posElement = new ElementTriplesBlock(posPart);
		ElementGroup combinedElementGroup = new ElementGroup();
		combinedElementGroup.addElement(posElement);
		for(BasicPattern negPart : negParts) {
			ElementGroup negGroup = new ElementGroup();
			negGroup.addElement(new ElementTriplesBlock(negPart));
			combinedElementGroup.addElementFilter(new ElementFilter(new E_NotExists(negGroup)));
		}
		Query query = new Query();
		query.setQuerySelectType();
		query.setQueryResultStar(true);
		query.setQueryPattern(combinedElementGroup);
		return query;
		
	}
	
	//////////////////////////////////////
	// I/O Methods (e.g., read queries from file)
	//////////////////////////////////////

	static String prefixes = "PREFIX wd: <http://www.wikidata.org/entity/>\r\nPREFIX xsd: <http://www.w3.org/2001/XMLSchema#>\r\nPREFIX schema: <http://schema.org/>\r\nPREFIX wikibase: <http://wikiba.se/ontology#>\r\nPREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>\r\n";
	
	public static List<Query> readQueries(String filename) {
		
		List<Query> queries = new ArrayList<Query>();
		try {
			String content = readFile(filename, StandardCharsets.UTF_8);
			String[] arr = content.split("##########");
			for(int i = 0; i < arr.length; i++) {
				queries.add(QueryFactory.create(prefixes+arr[i]));
			}
		} catch (IOException e) { e.printStackTrace(); }
		return queries;

	}
	
	// credit: http://stackoverflow.com/questions/326390/how-to-create-a-java-string-from-the-contents-of-a-file
	static String readFile(String path, Charset encoding) throws IOException {
		
	  byte[] encoded = Files.readAllBytes(Paths.get(path));
	  return new String(encoded, encoding);
	  
	}
	
	//////////////////////////////////////
	// Query-to-BGP and vice versa
	//////////////////////////////////////
	
	// get the body of a conjunctive query
	// assumption: q is a conjunctive SELECT query aka SELECT * WHERE { s1 p1 o1 . ... . sn pn on }
	public static BasicPattern body(Query q) {
		
		List<Triple> triples = new ArrayList<Triple>();
		PathBlock pb = ((ElementPathBlock) ((ElementGroup) q.getQueryPattern()).getElements().get(0)).getPattern(); // magic line :-)
		for(TriplePath tp : pb.getList()) {
			triples.add(tp.asTriple());
		}
		return BasicPattern.wrap(triples);
		
	}
	
	// construct a SELECT query from a triple pattern
	public static Query triplePatternToSelectQuery(Triple t) {
		
		List<Triple> bgp = new ArrayList<Triple>();
		bgp.add(t);
		return bgpToSelectQuery(BasicPattern.wrap(bgp));
		
	}
	
	// construct a SELECT query from a BGP
	public static Query bgpToSelectQuery(BasicPattern bgp) {
		
		ElementTriplesBlock element = new ElementTriplesBlock(bgp);
		Query query = new Query();
		query.setQuerySelectType();
		query.setQueryResultStar(true);
		query.setQueryPattern(element);
		return query;
		
	}
	
	// construct a CONSTRUCT query from a BGP
	public static Query bgpToConstructQuery(BasicPattern bgp) {
		
		Query query = new Query();
		query.setQueryConstructType();
		Template template = new Template(bgp);
	    query.setConstructTemplate(template);
	    Element element = new ElementTriplesBlock(bgp);
	    query.setQueryPattern(element);
		return query;
		
	}

	//////////////////////////////////////
	// Freeze/Defreeze aka Prototypical Graph Methods
	//////////////////////////////////////
	
	// namespace IRI for frozen variables
	static String frozenVarsNS  = "http://example.org/frozen-var/";

	public static Model freeze(BasicPattern bgp) {
		
		Model m = ModelFactory.createDefaultModel();
		for(Triple t : bgp)
			m.add(freeze(t));
		return m;
		
	}
	
	public static Statement freeze(Triple t) {
		
		Resource subject;
		Property predicate;
		RDFNode object;
		
		Node s = t.getSubject();
		Node p = t.getPredicate();
		Node o = t.getObject();

		if (s.isVariable())
			subject = ResourceFactory.createResource(frozenVarsNS + s.toString());
		else
			subject = ResourceFactory.createResource(s.toString());
		
		if (p.isVariable())
			predicate = ResourceFactory.createProperty(frozenVarsNS + p.toString());
		else
			predicate = ResourceFactory.createProperty(p.toString());
		
		if (o.isVariable())
			object = ResourceFactory.createResource(frozenVarsNS + o.toString());
		else if (o.isLiteral()) { // support for typed literas but not yet for lang tags
			if(o.getLiteralDatatypeURI() == null) // if no type
				object = ResourceFactory.createPlainLiteral(o.getLiteralLexicalForm());
			else			
				object = ResourceFactory.createTypedLiteral(o.getLiteralLexicalForm(), o.getLiteralDatatype());
		}
		else
			object = ResourceFactory.createResource(o.toString());
		
		return ResourceFactory.createStatement(subject, predicate, object);
		
	}
	
	// freeze a var
	// used in CompletenessTemplate.java
	public static Node freeze(Var v) {
		return NodeFactory.createURI(frozenVarsNS + v.toString());
	}
	
	// this will melt a frozen (or prototypical) graph to be a BGP again
 	public static BasicPattern defreeze(Model model) {
 		
 		List<Triple> tripleList = new ArrayList<Triple>();
 		StmtIterator it = model.listStatements();
 		while(it.hasNext()) {
 			Statement stmt = it.next();
 			Triple t = defreeze(stmt.asTriple());
 			tripleList.add(t);
 		}
 		return BasicPattern.wrap(tripleList);
 		
 	}
 	
 	public static Triple defreeze(Triple triple) {	
 		return Triple.create(defreeze(triple.getSubject()), defreeze(triple.getPredicate()), defreeze(triple.getObject()));
 	}
	
 	public static Node defreeze(Node n) {
 		
 		String nAsStr = n.toString();
 		if(n.isURI())
 			if(nAsStr.startsWith(frozenVarsNS))
 				n = Var.alloc(nAsStr.substring(frozenVarsNS.length() + 1));
 		return n;
 		
 	}

	//////////////////////////////////////
	// Check Equivalence (e.g., for removing redundant templates)
	//////////////////////////////////////
 	
 	public static List<Binding> areBGPsEquivalent(BasicPattern bgp1, BasicPattern bgp2) {
 		
 		List<Binding> containmentMapping1 = isContainedBy(bgp1, bgp2);
 		List<Binding> containmentMapping2 = isContainedBy(bgp2, bgp1);
 				
 		if(!containmentMapping1.isEmpty() && !containmentMapping2.isEmpty())
 			return containmentMapping1;
 		else
 			return new ArrayList<Binding>();
 		
 	}
 	
 	public static List<Binding> isContainedBy(BasicPattern bgp1, BasicPattern bgp2) {
 		
 		List<Binding> containmentMapping = new ArrayList<Binding>();
 		
 		Model frozenBGP1 = freeze(bgp1);
 		Query qBGP2 = bgpToSelectQuery(bgp2);
 		QueryExecution qExec = QueryExecutionFactory.create(qBGP2, frozenBGP1);
 		ResultSet rs = qExec.execSelect();
 		while(rs.hasNext())
 			containmentMapping.add(rs.nextBinding());
 		
 		return containmentMapping;
 		
 	}
 	
}
