package it.unibz.inf.soundness;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import it.unibz.inf.completeness.CompletenessReasoner;
import it.unibz.inf.completeness.CompletenessStatement;
import it.unibz.inf.completeness.CompletenessTemplate;
import it.unibz.inf.completeness.ConstantMap;
import it.unibz.inf.semweb.QueryToolkit;

import com.hp.hpl.jena.graph.Triple;
import com.hp.hpl.jena.query.Query;
import com.hp.hpl.jena.rdf.model.Model;
import com.hp.hpl.jena.rdf.model.ModelFactory;
import com.hp.hpl.jena.sparql.core.BasicPattern;
import com.hp.hpl.jena.sparql.core.Substitute;
import com.hp.hpl.jena.sparql.engine.binding.Binding;
import com.hp.hpl.jena.sparql.engine.binding.BindingFactory;
import com.hp.hpl.jena.sparql.engine.binding.BindingMap;

public class SoundnessReasoner {

	//////////////////////////////////////
	// Check Pattern Soundness
	//////////////////////////////////////
	
	public static boolean checkSchemaSoundnessNaive(BasicPattern posPart, List<BasicPattern> negParts, List<CompletenessStatement> statements, CompletenessReasoner completenessReasoner) {
		
		Model posPrototypicalGraph = QueryToolkit.freeze(posPart);
		for(BasicPattern negPart : negParts) {
			Model negPrototypicalGraph = QueryToolkit.freeze(negPart);
			Model posAndNegPrototypicalGraph = ModelFactory.createDefaultModel();
			posAndNegPrototypicalGraph.add(posPrototypicalGraph);
			posAndNegPrototypicalGraph.add(negPrototypicalGraph);
			if(!(completenessReasoner.checkConditionalSchemaCompletenessNaive(negPrototypicalGraph, posAndNegPrototypicalGraph, statements)))
				return false;
		}
		return true;
		
	}
	
	public static boolean checkSchemaSoundnessOptimized(BasicPattern posPart, List<BasicPattern> negParts, ConstantMap constantMap, CompletenessReasoner completenessReasoner) {
		
		Model posPrototypicalGraph = QueryToolkit.freeze(posPart);
		List<Triple> posPartAsList = posPart.getList();
		for(BasicPattern negPart : negParts) {
			Model negPrototypicalGraph = QueryToolkit.freeze(negPart);
			Model posAndNegPrototypicalGraph = ModelFactory.createDefaultModel();
			posAndNegPrototypicalGraph.add(posPrototypicalGraph);
			posAndNegPrototypicalGraph.add(negPrototypicalGraph);
			List<Triple> posAndNegPartAsList = new ArrayList<Triple>(posPartAsList);
			posAndNegPartAsList.addAll(negPart.getList());
			if(!(completenessReasoner.checkConditionalSchemaCompletenessOptimized(negPrototypicalGraph, posAndNegPartAsList, posAndNegPrototypicalGraph, constantMap)))
				return false;
		}
		return true;
		
	}

	//////////////////////////////////////
	// Check Answer Soundness
	//////////////////////////////////////
	
	public static List<Binding> getSoundAnswersNaive(BasicPattern posPart, List<BasicPattern> negParts, List<CompletenessStatement> statements, CompletenessReasoner reasoner) {

		List<Binding> soundAnswers = new ArrayList<Binding>();
		Query queryWithNegation = QueryToolkit.genQueryWithNegation(posPart, negParts);
		List<Binding> queryAnswers = reasoner.evalSELECT(queryWithNegation);
		
		for(Binding queryAnswer : queryAnswers) {
			boolean isSound = checkAnswerSoundnessNaive(queryAnswer, negParts, statements, reasoner);
			if(isSound)
				soundAnswers.add(queryAnswer);
		}
		return soundAnswers;
		
	}

	public static boolean checkAnswerSoundnessNaive(Binding mapping, List<BasicPattern> negParts, List<CompletenessStatement> statements, CompletenessReasoner reasoner) {
		for(BasicPattern negPart : negParts) {
			if(!(reasoner.checkCompletenessNaive(Substitute.substitute(negPart, mapping), statements))) {
				return false;
			}
		}
		return true;
	}

	public static List<Binding> getSoundAnswersNaiveFromBindings(List<Binding> queryAnswers, List<BasicPattern> negParts, List<CompletenessStatement> statements, CompletenessReasoner reasoner) {
		
		List<Binding> soundAnswers = new ArrayList<Binding>();
		
		for(Binding queryAnswer:queryAnswers) {
						
			boolean isSound = checkAnswerSoundnessNaive(queryAnswer, negParts, statements, reasoner);
			if(isSound) {
				soundAnswers.add(queryAnswer);
			}
			
		}
		return soundAnswers;
		
	}
	
	public static List<Binding> getSoundAnswersOptimizedFromBindings(List<Binding> queryAnswers, List<BasicPattern> negParts, HashMap<String, List<CompletenessTemplate>> templateMap, CompletenessReasoner reasoner) {

		// DEBUG
//		System.out.println("getSoundAnswersOptimizedFromBindings: " + negParts);
		
		List<Binding> soundAnswers = new ArrayList<Binding>();
		
		for(Binding queryAnswer:queryAnswers) {
			
			// DEBUG
//			System.out.println("Binding: " + queryAnswer);
			
			boolean isSound = checkAnswerSoundnessOptimized(queryAnswer, negParts, templateMap, reasoner);
			if(isSound) {
				soundAnswers.add(queryAnswer);
			}
			
			// DEBUG
//			reasoner.close();
//			System.exit(0);
			
		}
		return soundAnswers;
		
	}
	
	public static List<Binding> getSoundAnswersOptimized(BasicPattern posPart, List<BasicPattern> negParts, HashMap<String, List<CompletenessTemplate>> templateMap, CompletenessReasoner reasoner) {

		List<Binding> soundAnswers = new ArrayList<Binding>();
		Query queryWithNegation = QueryToolkit.genQueryWithNegation(posPart, negParts);
		List<Binding> queryAnswers = reasoner.evalSELECT(queryWithNegation);
		
		for(Binding queryAnswer : queryAnswers) {
			boolean isSound = checkAnswerSoundnessOptimized(queryAnswer, negParts, templateMap, reasoner);
			if(isSound) {
				soundAnswers.add(queryAnswer);
			}
		}
		return soundAnswers;
		
	}
	
	public static boolean checkAnswerSoundnessOptimized(Binding mapping, List<BasicPattern> negParts, HashMap<String, List<CompletenessTemplate>> templateMap, CompletenessReasoner reasoner) {

		for(BasicPattern negPart : negParts) {	
			if(!(reasoner.checkCompletenessOptimized(Substitute.substitute(negPart, mapping), templateMap))) {
				return false;
			}
		}
		return true;
		
	}

	public static List<Binding> getSoundAnswersPMatchOnlyFromBindings(List<Binding> queryAnswers, List<BasicPattern> negParts, HashMap<String, List<CompletenessTemplate>> templateMap, CompletenessReasoner reasoner) {
		
		List<Binding> soundAnswers = new ArrayList<Binding>();
				
		for(Binding queryAnswer:queryAnswers) {
			
			boolean isSound = checkAnswerSoundnessPMatchOnly(queryAnswer, negParts, templateMap, reasoner);
			if(isSound) {
				soundAnswers.add(queryAnswer);
			}
			
		}
		return soundAnswers;
		
	}

	public static boolean checkAnswerSoundnessPMatchOnly(Binding mapping, List<BasicPattern> negParts, HashMap<String, List<CompletenessTemplate>> templateMap, CompletenessReasoner reasoner) {
		
		for(BasicPattern negPart : negParts) {
			if(!(reasoner.checkCompletenessPMatchOnly(Substitute.substitute(negPart, mapping), templateMap))) {
				return false;
			}
		}
		return true;
		
	}
	
	public static List<CompletenessTemplate> transformIndividualStatementsToIndividualTemplates(
			List<CompletenessStatement> statements) {
		
		List<CompletenessTemplate> templates = new ArrayList<CompletenessTemplate>();
		
		Set<BindingMap> setWithEmptyBinding = new HashSet<BindingMap>();
		setWithEmptyBinding.add(BindingFactory.create());
		
		for(CompletenessStatement statement:statements) {
			CompletenessTemplate localTemplate = new CompletenessTemplate(statement.getBGP().getList(), setWithEmptyBinding);
			templates.add(localTemplate);
		}
		return templates;
	}
	
}
