package it.unibz.inf.soundness.experiment;

import java.nio.file.Files;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;

import com.hp.hpl.jena.sparql.engine.binding.Binding;

import it.unibz.inf.semweb.*;
import it.unibz.inf.soundness.*;
import it.unibz.inf.completeness.*;

// main class for running soundness experiments
public class SoundnessExperimentWithNaiveAndPMatchOnlyAnswerChecking {

	// for reliability, we take X repeats for each experiment observation
	static int observationRepeats = 10;

	static boolean warmingUpFlag = true;
	
	static int experimentCase = 3;
	
	static final int TIMEOUTTIME = 300;
	
	// default TDB directory
	static String tdbDir = "tdb-PREPROCESSED-wikidata-simple-statements-20160201";

	// queries file
	static String queriesFile = "queries-two.txt";

	// parameters: [observation repeats] [warming up] [case] [query filename]
	// [observation repeats]: how many should an observation is repeated, e.g., 1, 10, 20
	// [warming up]: should a warming up be performed? i.e., 0 (no), 1 (yes)
	// [case]: which case should be performed?
	//     1 : oneTP
	//     2 : oneTPoneTP
	//     3 : twoTPsTO
	//     4 : twoTPsSE
	// [query filename]: what is the query filename e.g., "test-queries.txt"
	public static void main(String[] args) {
		
		// credit: http://stackoverflow.com/questions/15571496/how-to-check-if-a-folder-exists
		if (!Files.isDirectory(Paths.get(tdbDir))) {
			tdbDir = "C:\\Users\\user\\Documents\\Datasets\\tdb-PREPROCESSED-wikidata-simple-statements-20160201";
		}
		
		int startFrom = 0;
		
		if(args.length > 0) {
			observationRepeats = Integer.parseInt(args[0]);
			if(Integer.parseInt(args[1]) == 0)
				warmingUpFlag = false;
			experimentCase = Integer.parseInt(args[2]);
			queriesFile = args[3];
			startFrom = Integer.parseInt(args[4]);
		}

	    Date date = new Date();
        System.out.println("Start: " + date.toString());
		System.out.println("##### OBSERVATION REPEAT(S): " + observationRepeats);
		System.out.println("##### WITH WARMING UP: " + warmingUpFlag);
		System.out.println("##### EXPERIMENT CASE: " + experimentCase);
		System.out.println("##### QUERIES FILENAME: " + queriesFile);
		System.out.println("##### TIMEOUT: " + TIMEOUTTIME);
		
		long startTime;
		long endTime;
		long elapsedTime;
		
		List<QueryWithNegation> queriesWithNegation;
		if(experimentCase == 1) {
			queriesWithNegation = QueryToolkit.genQueriesWithNegationOneTP(QueryToolkit.readQueries(queriesFile));
		}
		else if(experimentCase == 2) {
			queriesWithNegation = QueryToolkit.genQueriesWithNegationOneTPOneTP(QueryToolkit.readQueries(queriesFile));
		} else if(experimentCase == 5) {
			queriesWithNegation = QueryToolkit.genQueriesWithNegationThreeTPs(QueryToolkit.readQueries(queriesFile));
		}
		else {	
			queriesWithNegation = QueryToolkit.genQueriesWithNegationTwoTPs(QueryToolkit.readQueries(queriesFile));
		}
		
		System.out.println("##### QUERIES:\n" + queriesWithNegation + "\n");

		List<Observation> observations = new ArrayList<Observation>();
		for(QueryWithNegation queryWithNegation : queriesWithNegation) {
			observations.add(new Observation(queryWithNegation.toString(), queryWithNegation.getLength()));
		}
		
		// DEBUG
//		QueryWithNegation queryNeg = queriesWithNegation.get(0);
//		Model prototypicalGraph = QueryToolkit.freeze(queryNeg.posPart);
//		prototypicalGraph.add(QueryToolkit.freeze(queryNeg.negParts.get(0)));
//		System.out.println(prototypicalGraph);
//		System.out.println(QueryToolkit.defreeze(prototypicalGraph));
//		System.exit(0);
		
		CompletenessReasoner reasoner = new CompletenessReasoner(tdbDir);
		
		//////////////////////////////////////////////////
		// QUERY EVALUATION EXPERIMENTS
		//////////////////////////////////////////////////
		
		if(warmingUpFlag) {
		
			// WARMING UP 3 TIMES
			System.out.println("##### WARMING UP FOR QUERY EVALUATIONS:");
			for(QueryWithNegation queryWithNegation : queriesWithNegation) {
				System.out.println("Evaluating:\n" + queryWithNegation);
				reasoner.evalSELECT(queryWithNegation.getQuery());
				reasoner.evalSELECT(queryWithNegation.getQuery());
				reasoner.evalSELECT(queryWithNegation.getQuery());
			}
		
		}

		System.out.println("##### QUERY EVALUATIONS:");
		
		reasoner.close();
		for(int i = 0; i < observationRepeats; i++) {
			for(int j = 0; j < queriesWithNegation.size(); j++) {
				QueryWithNegation queryWithNegation = queriesWithNegation.get(j);
				
				// DEBUG
				System.out.println(queryWithNegation);
				
				reasoner = new CompletenessReasoner(tdbDir);
				startTime = System.nanoTime();
				List<Binding> queryEvalResults = reasoner.evalSELECT(queryWithNegation.getQuery());
				endTime = System.nanoTime();
				elapsedTime = endTime-startTime;
				int queryResultsSize = queryEvalResults.size();
				System.out.println(queryResultsSize + " results, with elapsed time: " + elapsedTime);
				observations.get(j).getQueryEvaluations().add(elapsedTime);
				observations.get(j).setQueryResultsSize(queryResultsSize);
				reasoner.close();
			}
			System.out.println("---");
		}
		reasoner = new CompletenessReasoner(tdbDir);
		System.out.println();
				
		//////////////////////////////////////////////////
		// ANSWER SOUNDNESS EXPERIMENTS
		//////////////////////////////////////////////////
		
		System.out.println("##### GENERATING COMPLETENESS TEMPLATES:");
		
		List<CompletenessTemplate> answerSoundnessTemplates;
		if(experimentCase != 4) {
			answerSoundnessTemplates = reasoner.genCompletenessTemplatesForAnswerSoundness(queriesWithNegation);
		} else {
			answerSoundnessTemplates = reasoner.genCompletenessTemplatesForAnswerSoundnessPerTriplePattern(queriesWithNegation);
		}
		
		System.out.println("\n##### GENERATED COMPLETENESS TEMPLATES: " + answerSoundnessTemplates.size() + "\n");

		answerSoundnessTemplates = CompletenessTemplate.removeRedundancies(answerSoundnessTemplates);
		System.out.println("##### NON-REDUNDANT TEMPLATES: " + answerSoundnessTemplates.size() + "\n");
		
		// DEBUG
//		for(CompletenessTemplate template : answerSoundnessTemplates) {
//			List<CompletenessStatement> statementsOfTemplate = template.unfold();
//			System.out.println(template + " : " + statementsOfTemplate.size() + " : " + template.bindings.size() + " : " + template.bindings);
//		}
//		reasoner.close();
//		System.exit(0);
		
		HashMap<String, List<CompletenessTemplate>> templateMap = SignatureFactory.getSignatureMapForTemplates(answerSoundnessTemplates);
		
		// naive answer soundness check
		List<CompletenessStatement> instantiatedStatements = new ArrayList<CompletenessStatement>();
		for(CompletenessTemplate template : answerSoundnessTemplates) {
			List<CompletenessStatement> statementsOfTemplate = template.unfold();
			instantiatedStatements.addAll(statementsOfTemplate);			
		}
		
		// free up memory
		answerSoundnessTemplates = new ArrayList<CompletenessTemplate>();
		
		HashMap<String, List<CompletenessTemplate>> templateMapForPMatch = SignatureFactory.getSignatureMapForTemplates(SoundnessReasoner.transformIndividualStatementsToIndividualTemplates(instantiatedStatements));
		
		if(warmingUpFlag) {
		
			// WARMING UP 3 TIMES
			System.out.println("##### WARMING UP ANSWER SOUNDNESS:\n");
			reasoner.close();
			for(QueryWithNegation queryWithNegation : queriesWithNegation) {
					reasoner = new CompletenessReasoner(tdbDir);
					System.out.println("Checking answer soundness of:\n" + queryWithNegation);
					SoundnessReasoner.getSoundAnswersOptimized(queryWithNegation.posPart, queryWithNegation.negParts, templateMap, reasoner);
					SoundnessReasoner.getSoundAnswersOptimized(queryWithNegation.posPart, queryWithNegation.negParts, templateMap, reasoner);
					SoundnessReasoner.getSoundAnswersOptimized(queryWithNegation.posPart, queryWithNegation.negParts, templateMap, reasoner);
					reasoner.close();
			}
			reasoner = new CompletenessReasoner(tdbDir);

		}
		
		System.out.println("##### ANSWER SOUNDNESS CHECKING:");
		
		reasoner.close();
		for(int i = 0; i < observationRepeats; i++) {
			
			List<String> pmatchCheckTimes = new ArrayList<String>();
			List<String> naiveCheckTimes = new ArrayList<String>();
			List<String> numOfAnswerss = new ArrayList<String>();
			
			for(int j = 0; j < queriesWithNegation.size(); j++) {
							
				QueryWithNegation queryWithNegation = queriesWithNegation.get(j);
				reasoner = new CompletenessReasoner(tdbDir);
				
				long queryEvalStartTime = System.nanoTime();
				List<Binding> answersToCheck = reasoner.evalSELECT(queryWithNegation.getQuery());
				long queryEvalEndTime = System.nanoTime();
				long queryEvalElapsedTime = queryEvalEndTime - queryEvalStartTime;
				
				long answerSoundnessCheckStartTime;
				long answerSoundnessCheckEndTime;
				long answerSoundnessCheckElapsedTime;
				
				numOfAnswerss.add(new Integer(answersToCheck.size()).toString());
				System.out.println("Nums so far: " + numOfAnswerss);
				
				int from = startFrom;
				if(j >= from) {
					
					System.out.println(j + " ID, RESULTS: " + answersToCheck.size());
					
//					 start pmatch only
					final Runnable stuffToDoPMatch = new MyThreadPMatch(answersToCheck, queryWithNegation, templateMapForPMatch, reasoner);
					final ExecutorService executorPMatch = Executors.newSingleThreadExecutor();
					final Future futurePMatch = executorPMatch.submit(stuffToDoPMatch);
					executorPMatch.shutdown(); // This does not cancel the already-scheduled task.
					
					try {
					  answerSoundnessCheckStartTime = System.nanoTime();
					  futurePMatch.get(TIMEOUTTIME, TimeUnit.SECONDS);
					  answerSoundnessCheckEndTime = System.nanoTime();
					  answerSoundnessCheckElapsedTime = answerSoundnessCheckEndTime - answerSoundnessCheckStartTime;
					  System.out.println("# (PMatch ONLY) Answer Soundness Check time: " + answerSoundnessCheckElapsedTime);
					  pmatchCheckTimes.add((new Long(answerSoundnessCheckElapsedTime)).toString());
					  System.out.println("PMATCH Times so far: " + pmatchCheckTimes);
					}
					catch (InterruptedException ie) { 
					  /* Handle the interruption. Or ignore it. */ 
					}
					catch (ExecutionException ee) { 
					  /* Handle the error. Or ignore it. */ 
					}
					catch (TimeoutException te) { 
					  System.out.println("# (PMATCH) Answer Soundness Check time: TIMEOUT!");
					  pmatchCheckTimes.add("TIMEOUT");
					  System.out.println("PMATCH Times so far: " + pmatchCheckTimes);
					}
					if (!executorPMatch.isTerminated())
					    executorPMatch.shutdownNow(); // If you want to stop the code that hasn't finished.
					// end pmatch only
			
				
//				// start pmatch+template
//				answerSoundnessCheckStartTime = System.nanoTime();
//				List<Binding> soundAnswers = SoundnessReasoner.getSoundAnswersOptimizedFromBindings(answersToCheck, queryWithNegation.negParts, templateMap, reasoner);
//				answerSoundnessCheckEndTime = System.nanoTime();
//				answerSoundnessCheckElapsedTime = answerSoundnessCheckEndTime - answerSoundnessCheckStartTime;
//				
//				long queryEvalAndAnswerSoundnessTime = queryEvalElapsedTime + answerSoundnessCheckElapsedTime;
//				
//				int soundAnswersSize = soundAnswers.size();
//		
//				System.out.println("#Sound answers: " + soundAnswersSize + ", with Query Evaluation time: " + queryEvalElapsedTime + ", with ANSWER SOUNDNESS CHECK TIME: " + answerSoundnessCheckElapsedTime + ", with total time: " + queryEvalAndAnswerSoundnessTime);
//				observations.get(j).getQueryEvaluationsInAnswerSoundnessChecks().add(queryEvalElapsedTime);
//				observations.get(j).getAnswerSoundnessChecksWithoutQueryEvaluation().add(answerSoundnessCheckElapsedTime);
//				observations.get(j).getAnswerSoundnessChecks().add(queryEvalAndAnswerSoundnessTime);
//				observations.get(j).setSoundAnswersSize(soundAnswersSize);
//				// end of pmatch+template
				
//				// start of naive
//				final Runnable stuffToDo = new MyThread(answersToCheck, queryWithNegation, instantiatedStatements, reasoner);
//				final ExecutorService executor = Executors.newSingleThreadExecutor();
//				final Future future = executor.submit(stuffToDo);
//				executor.shutdown(); // This does not cancel the already-scheduled task.
//				
//				try {
//				  answerSoundnessCheckStartTime = System.nanoTime();
//				  future.get(TIMEOUTTIME, TimeUnit.SECONDS);
//				  answerSoundnessCheckEndTime = System.nanoTime();
//				  answerSoundnessCheckElapsedTime = answerSoundnessCheckEndTime - answerSoundnessCheckStartTime;
//				  observations.get(j).getAnswerSoundnessChecksWithoutQueryEvaluationNaive().add(answerSoundnessCheckElapsedTime);
//				  System.out.println("# (NAIVE) Answer Soundness Check time: " + answerSoundnessCheckElapsedTime);
//				  naiveCheckTimes.add((new Long(answerSoundnessCheckElapsedTime)).toString());
//				  System.out.println("NAIVE Times so far: " + naiveCheckTimes);
//				}
//				catch (InterruptedException ie) { 
//				  /* Handle the interruption. Or ignore it. */ 
//				}
//				catch (ExecutionException ee) { 
//				  /* Handle the error. Or ignore it. */ 
//				}
//				catch (TimeoutException te) { 
//				  observations.get(j).getAnswerSoundnessChecksWithoutQueryEvaluationNaive().add(new Long(99999999));
//				  System.out.println("# (NAIVE) Answer Soundness Check time: TIMEOUT!");
//				  naiveCheckTimes.add("TIMEOUT");
//				  System.out.println("NAIVE Times so far: " + naiveCheckTimes);
//				}
//				if (!executor.isTerminated())
//				    executor.shutdownNow(); // If you want to stop the code that hasn't finished.
//				// end of naive
				
				}
				
				reasoner.close();
			}
			System.out.println("---");
		}
		reasoner = new CompletenessReasoner(tdbDir);
		System.out.println();
		
		// DEBUG
//		Observation.printAll(observations);
//		reasoner.close();
//		System.exit(0);
		
		//////////////////////////////////////////////////
		// PATTERN SOUNDNESS EXPERIMENTS
		//////////////////////////////////////////////////
		
		// naive answer soundness check
		instantiatedStatements = new ArrayList<CompletenessStatement>();
		
		// free memory
		templateMap = new HashMap<String, List<CompletenessTemplate>>();
		
		System.out.println("##### GENERATING COMPLETENESS STATEMENTS:");
		
		List<CompletenessTemplate> patternSoundnessTemplates = new ArrayList<CompletenessTemplate>();
				
		if(experimentCase != 4) {
			answerSoundnessTemplates = reasoner.genCompletenessTemplatesForAnswerSoundness(queriesWithNegation);
		} else {
			answerSoundnessTemplates = reasoner.genCompletenessTemplatesForAnswerSoundnessPerTriplePattern(queriesWithNegation);
		}
			
		answerSoundnessTemplates = CompletenessTemplate.removeRedundancies(answerSoundnessTemplates);		

		patternSoundnessTemplates.addAll(answerSoundnessTemplates);
		patternSoundnessTemplates.addAll(reasoner.genCompletenessTemplatesForPatternSoundness(queriesWithNegation));

		// DEBUG
//		for(int i = 0; i < 1000; i++)
//			patternSoundnessTemplates.addAll(reasoner.genCompletenessTemplatesForPatternSoundness(queriesWithNegation));

		List<CompletenessStatement> statements = new ArrayList<CompletenessStatement>();
		for(CompletenessTemplate template : patternSoundnessTemplates) {
			List<CompletenessStatement> statementsOfTemplate = template.unfold();
			statements.addAll(statementsOfTemplate);
			
			// DEBUG
//			System.out.println(statementsOfTemplate.size() + " : " + template.bindings.size());
			
		}
		
		// free memory
		patternSoundnessTemplates = new ArrayList<CompletenessTemplate>();
		
		System.out.println("\n##### GENERATED COMPLETENESS STATEMENTS FOR PATTERN SOUNDNESS: " + statements.size() + "\n");
		ConstantMap constantMap = new ConstantMap(statements);
		
		// naive pattern soundness checking
//		statements = new ArrayList<CompletenessStatement>(); // free up some memory

		if(warmingUpFlag) {
		
			// WARMING UP 3 TIMES
			System.out.println("##### WARMING UP PATTERN SOUNDNESS:\n");
			reasoner.close();
			for(QueryWithNegation queryWithNegation : queriesWithNegation) {
				reasoner = new CompletenessReasoner(tdbDir);
				System.out.println("Checking pattern soundness of:\n" + queryWithNegation);
				SoundnessReasoner.checkSchemaSoundnessOptimized(queryWithNegation.posPart, queryWithNegation.negParts, constantMap, reasoner);
				SoundnessReasoner.checkSchemaSoundnessOptimized(queryWithNegation.posPart, queryWithNegation.negParts, constantMap, reasoner);
				SoundnessReasoner.checkSchemaSoundnessOptimized(queryWithNegation.posPart, queryWithNegation.negParts, constantMap, reasoner);
				reasoner.close();
			}
			reasoner = new CompletenessReasoner(tdbDir);

		}
		
		System.out.println("##### PATTERN SOUNDNESS CHECKING:");
		reasoner.close();
		for(int i = 0; i < observationRepeats; i++) {
			for(int j = 0; j < queriesWithNegation.size(); j++) {
				QueryWithNegation queryWithNegation = queriesWithNegation.get(j);
				reasoner = new CompletenessReasoner(tdbDir);
				startTime = System.nanoTime();
				boolean isSound = SoundnessReasoner.checkSchemaSoundnessOptimized(queryWithNegation.posPart, queryWithNegation.negParts, constantMap, reasoner);
				endTime = System.nanoTime();
				elapsedTime = endTime-startTime;
				System.out.println("#Sound graph pattern: " + isSound + ", with time elapsed " + elapsedTime);
				observations.get(j).setPatternSoundnessFlag(isSound);
				observations.get(j).getPatternSoundnessChecks().add(elapsedTime);
				
				// naive pattern soundness checking
				startTime = System.nanoTime();
				isSound = SoundnessReasoner.checkSchemaSoundnessNaive(queryWithNegation.posPart, queryWithNegation.negParts, statements, reasoner);
				endTime = System.nanoTime();
				elapsedTime = endTime-startTime;
				System.out.println("# (NAIVE) Sound graph pattern: " + isSound + ", with time elapsed " + elapsedTime);
				observations.get(j).getPatternSoundnessChecksNaive().add(elapsedTime);
				
				reasoner.close();
			}
			System.out.println("---");
		}
		reasoner = new CompletenessReasoner(tdbDir);
		System.out.println();
		
		System.out.println("##### STATISTICS:");
		Observation.printAll(observations);
		
		date = new Date();
        System.out.println("Finish: " + date.toString());
		
		reasoner.close();
	}
	
}
